@@ -24,6 +24,7 @@
 #include "qgsrendercontext.h"
 #include "qgslayoutundocommand.h"
 #include "qgslayoutmeasurement.h"
+#include "qgsapplication.h"
 #include <QGraphicsRectItem>
 #include <QIcon>
 #include <QPainter>
@@ -32,7 +33,7 @@ class QgsLayout;
 class QPainter;
 class QgsLayoutItemGroup;
 class QgsLayoutEffect;
-
+class QgsStyleEntityVisitorInterface;
 
 /**
  * \ingroup core
@@ -102,7 +103,6 @@ class CORE_EXPORT QgsLayoutItemRenderContext
     double mViewScaleFactor = 1.0;
 };
 
-
 /**
  * \ingroup core
  * \class QgsLayoutItem
@@ -125,9 +125,14 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
 #include "qgslayoutitempage.h"
 #endif
 
-
 #ifdef SIP_RUN
     SIP_CONVERT_TO_SUBCLASS_CODE
+
+    // FREAKKKKIIN IMPORTANT!!!!!!!!!!!
+    // IF YOU PUT SOMETHING HERE, PUT IT IN QgsLayoutObject CASTING *****ALSO******
+    // (it's not enough for it to be in only one of the places, as sip inconsistently
+    // decides which casting code to perform here)
+
     // the conversions have to be static, because they're using multiple inheritance
     // (seen in PyQt4 .sip files for some QGraphicsItem classes)
     switch ( sipCpp->type() )
@@ -177,12 +182,16 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
         sipType = sipType_QgsLayoutFrame;
         *sipCppRet = static_cast<QgsLayoutFrame *>( sipCpp );
         break;
+
+      // did you read that comment above? NO? Go read it now. You're about to break stuff.
+
       default:
-        sipType = 0;
+        sipType = NULL;
     }
     SIP_END
 #endif
 
+
     Q_OBJECT
     Q_PROPERTY( bool locked READ isLocked WRITE setLocked NOTIFY lockChanged )
 
@@ -228,6 +237,8 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
       UndoMapGridAnnotationFontColor, //!< Map frame annotation color
       UndoMapGridLineSymbol, //!< Grid line symbol
       UndoMapGridMarkerSymbol, //!< Grid marker symbol
+      UndoMapGridIntervalRange, //!< Grid interval range
+      UndoMapLabelMargin, //!< Margin for labels from edge of map
       UndoPictureRotation, //!< Picture rotation
       UndoPictureFillColor, //!< Picture fill color
       UndoPictureStrokeColor, //!< Picture stroke color
@@ -263,6 +274,8 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
       UndoScaleBarSegmentsLeft, //!< Scalebar segments left
       UndoScaleBarSegments, //!< Scalebar number of segments
       UndoScaleBarHeight, //!< Scalebar height
+      UndoScaleBarSubdivisions, //!< Scalebar number of subdivisions
+      UndoScaleBarSubdivisionsHeight, //!< Scalebar subdivisions height
       UndoScaleBarFontColor, //!< Scalebar font color
       UndoScaleBarFillColor, //!< Scalebar fill color
       UndoScaleBarFillColor2, //!< Scalebar secondary fill color
@@ -279,10 +292,20 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
       UndoCustomCommand, //!< Base id for plugin based item undo commands
     };
 
+    /**
+     * Flags for controlling how an item behaves.
+     * \since QGIS 3.4.3
+     */
+    enum Flag
+    {
+      FlagOverridesPaint = 1 << 1,  //!< Item overrides the default layout item painting method
+    };
+    Q_DECLARE_FLAGS( Flags, Flag )
+
     /**
      * Constructor for QgsLayoutItem, with the specified parent \a layout.
      *
-     * If \a manageZValue is true, the z-Value of this item will be managed by the layout.
+     * If \a manageZValue is TRUE, the z-Value of this item will be managed by the layout.
      * Generally this is the desired behavior.
      */
     explicit QgsLayoutItem( QgsLayout *layout, bool manageZValue = true );
@@ -316,6 +339,12 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
     */
     virtual QString uuid() const { return mUuid; }
 
+    /**
+     * Returns the item's flags, which indicate how the item behaves.
+     * \since QGIS 3.4.3
+     */
+    virtual Flags itemFlags() const;
+
     /**
      * Returns the item's ID name. This is not necessarily unique, and duplicate ID names may exist
      * for a layout.
@@ -361,14 +390,14 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
     void setLocked( bool locked );
 
     /**
-     * Returns true if the item is locked, and cannot be interacted with using the mouse.
+     * Returns TRUE if the item is locked, and cannot be interacted with using the mouse.
      * \see setLocked()
      * \see lockChanged()
      */
     bool isLocked() const { return mIsLocked; }
 
     /**
-     * Returns true if the item is part of a QgsLayoutItemGroup group.
+     * Returns TRUE if the item is part of a QgsLayoutItemGroup group.
      * \see parentGroup()
      * \see setParentGroup()
      */
@@ -388,6 +417,26 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
      */
     void setParentGroup( QgsLayoutItemGroup *group );
 
+    /**
+     * Behavior of item when exporting to layered outputs.
+     * \since QGIS 3.10
+     */
+    enum ExportLayerBehavior
+    {
+      CanGroupWithAnyOtherItem, //!< Item can be placed on a layer with any other item (default behavior)
+      CanGroupWithItemsOfSameType, //!< Item can only be placed on layers with other items of the same type, but multiple items of this type can be grouped together
+      MustPlaceInOwnLayer, //!< Item must be placed in its own individual layer
+      ItemContainsSubLayers, //!< Item contains multiple sublayers which must be individually exported
+    };
+
+    /**
+     * Returns the behavior of this item during exporting to layered exports (e.g. SVG).
+     * \see numberExportLayers()
+     * \see exportLayerDetails()
+     * \since QGIS 3.10
+     */
+    virtual ExportLayerBehavior exportLayerBehavior() const;
+
     /**
      * Returns the number of layers that this item requires for exporting during layered exports (e.g. SVG).
      * Returns 0 if this item is to be placed on the same layer as the previous item,
@@ -395,8 +444,78 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
      *
      * Items which require multiply layers should check QgsLayoutContext::currentExportLayer() during
      * their rendering to determine which layer should be drawn.
+     *
+     * \see exportLayerBehavior()
+     * \see exportLayerDetails()
+     *
+     * \deprecated Use nextExportPart() and exportLayerBehavior() instead.
      */
-    virtual int numberExportLayers() const { return 0; }
+    Q_DECL_DEPRECATED virtual int numberExportLayers() const SIP_DEPRECATED;
+
+    /**
+     * Starts a multi-layer export operation.
+     *
+     * \see stopLayeredExport()
+     * \see nextExportPart()
+     * \since QGIS 3.10
+     */
+    virtual void startLayeredExport();
+
+    /**
+     * Stops a multi-layer export operation.
+     *
+     * \see startLayeredExport()
+     * \see nextExportPart()
+     * \since QGIS 3.10
+     */
+    virtual void stopLayeredExport();
+
+    /**
+     * Moves to the next export part for a multi-layered export item, during a multi-layered export.
+     *
+     * \see startLayeredExport()
+     * \see stopLayeredExport()
+     * \since QGIS 3.10
+     */
+    virtual bool nextExportPart();
+
+    /**
+     * Contains details of a particular export layer relating to a layout item.
+     * \ingroup core
+     * \since QGIS 3.10
+     */
+    struct CORE_EXPORT ExportLayerDetail
+    {
+      //! User-friendly name for the export layer
+      QString name;
+
+      //! Associated map layer ID, or an empty string if this export layer is not associated with a map layer
+      QString mapLayerId;
+
+      /**
+       * Associated composition mode if this layer is associated with a map layer
+       * \since QGIS 3.14
+       */
+      QPainter::CompositionMode compositionMode = QPainter::CompositionMode_SourceOver;
+
+      /**
+       * Associated opacity, if this layer is associated with a map layer
+       * \since QGIS 3.14
+       */
+      double opacity = 1.0;
+
+      //! Associated map theme, or an empty string if this export layer does not need to be associated with a map theme
+      QString mapTheme;
+    };
+
+    /**
+     * Returns the details for the specified current export layer.
+     *
+     * Only valid between calls to startLayeredExport() and stopLayeredExport()
+     *
+     * \since QGIS 3.10
+     */
+    virtual QgsLayoutItem::ExportLayerDetail exportLayerDetails() const;
 
     /**
      * Handles preparing a paint surface for the layout item and painting the item's
@@ -411,7 +530,7 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
      * so that the position of the reference point within the layout remains unchanged.
      * \see referencePoint()
      */
-    void setReferencePoint( const ReferencePoint &point );
+    void setReferencePoint( ReferencePoint point );
 
     /**
      * Returns the reference point for positioning of the layout item. This point is also
@@ -443,7 +562,7 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
      * size will place restrictions on the allowed item size. Data defined item size overrides
      * will also override the specified target size.
      *
-     * If \a includesFrame is true, then the size specified by \a size includes the
+     * If \a includesFrame is TRUE, then the size specified by \a size includes the
      * item's frame.
      *
      * \see minimumSize()
@@ -456,14 +575,14 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
     /**
      * Attempts to move the item to a specified \a point.
      *
-     * If \a useReferencePoint is true, this method will respect the item's
+     * If \a useReferencePoint is TRUE, this method will respect the item's
      * reference point, in that the item will be moved so that its current reference
      * point is placed at the specified target point.
      *
-     * If \a useReferencePoint is false, the item will be moved so that \a point
+     * If \a useReferencePoint is FALSE, the item will be moved so that \a point
      * falls at the top-left corner of the item.
      *
-     * If \a includesFrame is true, then the position specified by \a point represents the
+     * If \a includesFrame is TRUE, then the position specified by \a point represents the
      * point at which to place the outside of the item's frame.
      *
      * If \a page is not left at the default -1 value, then the position specified by \a point
@@ -484,7 +603,7 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
      * Attempts to update the item's position and size to match the passed \a rect in layout
      * coordinates.
      *
-     * If \a includesFrame is true, then the position and size specified by \a rect represents the
+     * If \a includesFrame is TRUE, then the position and size specified by \a rect represents the
      * position and size at for the outside of the item's frame.
      *
      * Note that the final position and size of the item may not match the specified target rect,
@@ -599,7 +718,7 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
     QgsAbstractLayoutUndoCommand *createCommand( const QString &text, int id, QUndoCommand *parent = nullptr ) override SIP_FACTORY;
 
     /**
-     * Returns true if the item includes a frame.
+     * Returns TRUE if the item includes a frame.
      * \see setFrameEnabled()
      * \see frameStrokeWidth()
      * \see frameJoinStyle()
@@ -626,7 +745,7 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
     void setFrameStrokeColor( const QColor &color );
 
     /**
-     * Returns the frame's stroke color. This is only used if frameEnabled() returns true.
+     * Returns the frame's stroke color. This is only used if frameEnabled() returns TRUE.
      * \see frameEnabled()
      * \see setFrameStrokeColor()
      * \see frameJoinStyle()
@@ -641,10 +760,10 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
      * \see setFrameJoinStyle()
      * \see setFrameStrokeColor()
      */
-    virtual void setFrameStrokeWidth( const QgsLayoutMeasurement &width );
+    virtual void setFrameStrokeWidth( QgsLayoutMeasurement width );
 
     /**
-     * Returns the frame's stroke width. This is only used if frameEnabled() returns true.
+     * Returns the frame's stroke width. This is only used if frameEnabled() returns TRUE.
      * \see frameEnabled()
      * \see setFrameStrokeWidth()
      * \see frameJoinStyle()
@@ -671,7 +790,7 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
     void setFrameJoinStyle( Qt::PenJoinStyle style );
 
     /**
-     * Returns true if the item has a background.
+     * Returns TRUE if the item has a background.
      * \see setBackgroundEnabled()
      * \see backgroundColor()
      */
@@ -686,7 +805,7 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
 
     /**
      * Returns the background color for this item. This is only used if hasBackground()
-     * returns true.
+     * returns TRUE.
      * \see setBackgroundColor()
      * \see hasBackground()
      */
@@ -742,7 +861,7 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
     void setExcludeFromExports( bool exclude );
 
     /**
-     * Returns true if the item contains contents with blend modes or transparency
+     * Returns TRUE if the item contains contents with blend modes or transparency
      * effects which can only be reproduced by rastering the item.
      *
      * Subclasses should ensure that implemented overrides of this method
@@ -753,7 +872,7 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
     virtual bool containsAdvancedEffects() const;
 
     /**
-     * Returns true if the item is drawn in such a way that forces the whole layout
+     * Returns TRUE if the item is drawn in such a way that forces the whole layout
      * to be rasterized when exporting to vector formats.
      * \see containsAdvancedEffects()
      */
@@ -837,6 +956,17 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
 
     QgsExpressionContext createExpressionContext() const override;
 
+    /**
+     * Accepts the specified style entity \a visitor, causing it to visit all style entities associated
+     * with the layout item.
+     *
+     * Returns TRUE if the visitor should continue visiting other objects, or FALSE if visiting
+     * should be canceled.
+     *
+     * \since QGIS 3.10
+     */
+    virtual bool accept( QgsStyleEntityVisitorInterface *visitor ) const;
+
   public slots:
 
     /**
@@ -866,8 +996,8 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
     /**
      * Sets the layout item's \a rotation, in degrees clockwise.
      *
-     * If \a adjustPosition is true, then this rotation occurs around the center of the item.
-     * If \a adjustPosition is false, rotation occurs around the item origin.
+     * If \a adjustPosition is TRUE, then this rotation occurs around the center of the item.
+     * If \a adjustPosition is FALSE, rotation occurs around the item origin.
      *
      * \see itemRotation()
      * \see rotateItem()
@@ -879,7 +1009,7 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
      * \see setItemRotation()
      * \see itemRotation()
     */
-    virtual void rotateItem( double angle, const QPointF &transformOrigin );
+    virtual void rotateItem( double angle, QPointF transformOrigin );
 
   signals:
 
@@ -905,6 +1035,13 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
      */
     void sizePositionChanged();
 
+    /**
+     * Emitted whenever the number of background tasks an item is executing changes.
+     *
+     * \since QGIS 3.10
+     */
+    void backgroundTaskCountChanged( int count );
+
   protected:
 
     /**
@@ -956,7 +1093,7 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
      * \see setFixedSize()
      * \see setMinimumSize()
      */
-    virtual QSizeF applyItemSizeConstraint( const QSizeF &targetSize );
+    virtual QSizeF applyItemSizeConstraint( QSizeF targetSize );
 
     /**
      * Refreshes an item's size by rechecking it against any possible item fixed
@@ -988,21 +1125,21 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
 
     /**
      * Refresh item's opacity, considering data defined opacity.
-      * If \a updateItem is set to false the item will not be automatically
+      * If \a updateItem is set to FALSE the item will not be automatically
       * updated after the opacity is set and a later call to update() must be made.
      */
     void refreshOpacity( bool updateItem = true );
 
     /**
      * Refresh item's frame, considering data defined colors and frame size.
-     * If \a updateItem is set to false, the item will not be automatically updated
+     * If \a updateItem is set to FALSE, the item will not be automatically updated
      * after the frame is set and a later call to update() must be made.
      */
     void refreshFrame( bool updateItem = true );
 
     /**
      * Refresh item's background color, considering data defined colors.
-     * If \a updateItem is set to false, the item will not be automatically updated
+     * If \a updateItem is set to FALSE, the item will not be automatically updated
      * after the frame color is set and a later call to update() must be made.
      */
     void refreshBackgroundColor( bool updateItem = true );
@@ -1016,12 +1153,12 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
      * Adjusts the specified \a point at which a \a reference position of the item
      * sits and returns the top left corner of the item, if reference point were placed at the specified position.
      */
-    QPointF adjustPointForReferencePosition( const QPointF &point, const QSizeF &size, const ReferencePoint &reference ) const;
+    QPointF adjustPointForReferencePosition( QPointF point, QSizeF size, ReferencePoint reference ) const;
 
     /**
      * Returns the current position (in layout units) of a \a reference point for the item.
     */
-    QPointF positionAtReferencePoint( const ReferencePoint &reference ) const;
+    QPointF positionAtReferencePoint( ReferencePoint reference ) const;
 
     /**
      * Returns the position for the reference point of the item, if the top-left of the item
@@ -1101,6 +1238,7 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
 
     //! Item opacity, between 0 and 1
     double mOpacity = 1.0;
+    double mEvaluatedOpacity = 1.0;
 
     QImage mItemCachedImage;
     double mItemCacheDpi = -1;
@@ -1129,15 +1267,14 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
     void preparePainter( QPainter *painter );
     bool shouldDrawAntialiased() const;
     bool shouldDrawDebugRect() const;
-
-    QSizeF applyMinimumSize( const QSizeF &targetSize );
-    QSizeF applyFixedSize( const QSizeF &targetSize );
+    QSizeF applyMinimumSize( QSizeF targetSize );
+    QSizeF applyFixedSize( QSizeF targetSize );
     QgsLayoutPoint applyDataDefinedPosition( const QgsLayoutPoint &position );
 
     double applyDataDefinedRotation( double rotation );
     void updateStoredItemPosition();
-    QPointF itemPositionAtReferencePoint( ReferencePoint reference, const QSizeF &size ) const;
-    void setScenePos( const QPointF &destinationPos );
+    QPointF itemPositionAtReferencePoint( ReferencePoint reference, QSizeF size ) const;
+    void setScenePos( QPointF destinationPos );
     bool shouldBlockUndoCommands() const;
 
     void applyDataDefinedOrientation( double &width, double &height, const QgsExpressionContext &context );
@@ -1149,6 +1286,8 @@ class CORE_EXPORT QgsLayoutItem : public QgsLayoutObject, public QGraphicsRectIt
     friend class QgsCompositionConverter;
 };
 
+Q_DECLARE_OPERATORS_FOR_FLAGS( QgsLayoutItem::Flags )
+
 #endif //QGSLAYOUTITEM_H
 
 
