@@ -15,13 +15,15 @@
  *                                                                         *
  ***************************************************************************/
 
-#ifndef QGSPOLYGONV2_H
-#define QGSPOLYGONV2_H
+#ifndef QGSPOLYGON_H
+#define QGSPOLYGON_H
 
 #include "qgis_core.h"
-#include "qgis.h"
+#include "qgis_sip.h"
 #include "qgscurvepolygon.h"
 
+class QgsLineString;
+
 /**
  * \ingroup core
  * \class QgsPolygon
@@ -33,11 +35,20 @@ class CORE_EXPORT QgsPolygon: public QgsCurvePolygon
   public:
     QgsPolygon();
 
+    /**
+     * Constructor for QgsPolygon, with the specified \a exterior ring and interior \a rings.
+     *
+     * Ownership of \a exterior and \a rings is transferred to the polygon.
+     *
+     * \since QGIS 3.14
+     */
+    QgsPolygon( QgsLineString *exterior SIP_TRANSFER, const QList< QgsLineString * > &rings SIP_TRANSFER = QList< QgsLineString * >() );
+
     QString geometryType() const override;
     QgsPolygon *clone() const override SIP_FACTORY;
     void clear() override;
     bool fromWkb( QgsConstWkbPtr &wkb ) override;
-    QByteArray asWkb() const override;
+    QByteArray asWkb( QgsAbstractGeometry::WkbFlags flags = QgsAbstractGeometry::WkbFlags() ) const override;
     QgsPolygon *surfaceToPolygon() const override SIP_FACTORY;
 
     /**
@@ -100,4 +111,4 @@ class CORE_EXPORT QgsPolygon: public QgsCurvePolygon
     friend class QgsCurvePolygon;
 
 };
-#endif // QGSPOLYGONV2_H
+#endif // QGSPOLYGON_H
