@@ -21,6 +21,7 @@
 #include "qgis_core.h"
 #include "qgis_sip.h"
 #include <QColor>
+#include <QPainter>
 #include <memory>
 
 #include "qgscoordinatetransform.h"
@@ -33,11 +34,17 @@
 #include "qgsdistancearea.h"
 #include "qgscoordinatetransformcontext.h"
 #include "qgspathresolver.h"
+#include "qgssymbollayerreference.h"
+#include "qgstemporalrangeobject.h"
 
 class QPainter;
 class QgsAbstractGeometry;
 class QgsLabelingEngine;
 class QgsMapSettings;
+class QgsRenderedFeatureHandlerInterface;
+class QgsSymbolLayer;
+
+class QgsMaskIdProvider;
 
 
 /**
@@ -46,8 +53,8 @@ class QgsMapSettings;
  * The context of a rendering operation defines properties such as
  * the conversion ratio between screen and map units, the extents
  * to be rendered etc.
- **/
-class CORE_EXPORT QgsRenderContext
+ */
+class CORE_EXPORT QgsRenderContext : public QgsTemporalRangeObject
 {
   public:
     QgsRenderContext();
@@ -71,6 +78,8 @@ class CORE_EXPORT QgsRenderContext
       Antialiasing             = 0x80,  //!< Use antialiasing while drawing
       RenderPartialOutput      = 0x100, //!< Whether to make extra effort to update map image with partially rendered layers (better for interactive map canvas). Added in QGIS 3.0
       RenderPreviewJob         = 0x200, //!< Render is a 'canvas preview' render, and shortcuts should be taken to ensure fast rendering
+      RenderBlocking           = 0x400, //!< Render and load remote sources in the same thread to ensure rendering remote sources (svg and images). WARNING: this flag must NEVER be used from GUI based applications (like the main QGIS application) or crashes will result. Only for use in external scripts or QGIS server.
+      RenderSymbolPreview      = 0x800, //!< The render is for a symbol preview only and map based properties may not be available, so care should be taken to handle map unit based sizes in an appropriate way.
     };
     Q_DECLARE_FLAGS( Flags, Flag )
 
@@ -164,6 +173,49 @@ class CORE_EXPORT QgsRenderContext
     */
     QPainter *painter() {return mPainter;}
 
+#ifndef SIP_RUN
+
+    /**
+     * Returns the const destination QPainter for the render operation.
+     * \see setPainter()
+    * \since QGIS 3.12
+    */
+    const QPainter *painter() const { return mPainter; }
+#endif
+
+    /**
+     * Returns a mask QPainter for the render operation.
+     * Multiple mask painters can be defined, each with a unique identifier.
+     * nullptr is returned if a mask painter with the given identifier does not exist.
+     * This is currently used to implement selective masking.
+     * \see setMaskPainter()
+     * \see currentMaskId()
+     * \since QGIS 3.12
+     */
+    QPainter *maskPainter( int id = 0 ) { return mMaskPainter.value( id, nullptr ); }
+
+    /**
+     * When rendering a map layer in a second pass (for selective masking),
+     * some symbol layers may be disabled.
+     *
+     * Returns the list of disabled symbol layers.
+     * \see setDisabledSymbolLayers()
+     * \see isSymbolLayerEnabled()
+     * \since QGIS 3.12
+     */
+    QSet<const QgsSymbolLayer *> disabledSymbolLayers() const { return mDisabledSymbolLayers; }
+
+    /**
+     * When rendering a map layer in a second pass (for selective masking),
+     * some symbol layers may be disabled.
+     *
+     * Checks whether a given symbol layer has been disabled for the current pass.
+     * \see setDisabledSymbolLayers()
+     * \see disabledSymbolLayers()
+     * \since QGIS 3.12
+     */
+    bool isSymbolLayerEnabled( const QgsSymbolLayer *layer ) const { return ! mDisabledSymbolLayers.contains( layer ); }
+
     /**
      * Returns the current coordinate transform for the context.
      *
@@ -249,6 +301,11 @@ class CORE_EXPORT QgsRenderContext
      */
     QgsRectangle mapExtent() const { return mOriginalMapExtent; }
 
+    /**
+     * Returns the context's map to pixel transform, which transforms between map coordinates and device coordinates.
+     *
+     * \see setMapToPixel()
+     */
     const QgsMapToPixel &mapToPixel() const {return mMapToPixel;}
 
     /**
@@ -259,20 +316,41 @@ class CORE_EXPORT QgsRenderContext
      */
     double scaleFactor() const {return mScaleFactor;}
 
+    /**
+     * Returns TRUE if the rendering operation has been stopped and any ongoing
+     * rendering should be canceled immediately.
+     *
+     * \see setRenderingStopped()
+     */
     bool renderingStopped() const {return mRenderingStopped;}
 
+    /**
+     * Returns TRUE if rendering operations should use vector operations instead
+     * of any faster raster shortcuts.
+     *
+     * \see setForceVectorOutput()
+     */
     bool forceVectorOutput() const;
 
     /**
-     * Returns true if advanced effects such as blend modes such be used
+     * Returns TRUE if advanced effects such as blend modes such be used
+     *
+     * \see setUseAdvancedEffects()
      */
     bool useAdvancedEffects() const;
 
     /**
      * Used to enable or disable advanced effects such as blend modes
+     *
+     * \see useAdvancedEffects()
      */
     void setUseAdvancedEffects( bool enabled );
 
+    /**
+     * Returns TRUE if edit markers should be drawn during the render operation.
+     *
+     * \see setDrawEditingInformation()
+     */
     bool drawEditingInformation() const;
 
     /**
@@ -283,16 +361,21 @@ class CORE_EXPORT QgsRenderContext
     double rendererScale() const {return mRendererScale;}
 
     /**
-     * Gets access to new labeling engine (may be nullptr)
+     * Gets access to new labeling engine (may be NULLPTR)
      * \note not available in Python bindings
      */
     QgsLabelingEngine *labelingEngine() const { return mLabelingEngine; } SIP_SKIP
 
+    /**
+     * Returns the color to use when rendering selected features.
+     *
+     * \see setSelectionColor()
+     */
     QColor selectionColor() const { return mSelectionColor; }
 
     /**
-     * Returns true if vector selections should be shown in the rendered map
-     * \returns true if selections should be shown
+     * Returns TRUE if vector selections should be shown in the rendered map
+     * \returns TRUE if selections should be shown
      * \see setShowSelection
      * \see selectionColor
      * \since QGIS v2.4
@@ -313,6 +396,11 @@ class CORE_EXPORT QgsRenderContext
      */
     void setCoordinateTransform( const QgsCoordinateTransform &t );
 
+    /**
+     * Sets the context's map to pixel transform, which transforms between map coordinates and device coordinates.
+     *
+     * \see mapToPixel()
+     */
     void setMapToPixel( const QgsMapToPixel &mtp ) {mMapToPixel = mtp;}
 
     /**
@@ -341,8 +429,19 @@ class CORE_EXPORT QgsRenderContext
      */
     void setMapExtent( const QgsRectangle &extent ) { mOriginalMapExtent = extent; }
 
+    /**
+     * Sets whether edit markers should be drawn during the render operation.
+     *
+     * \see drawEditingInformation()
+     */
     void setDrawEditingInformation( bool b );
 
+    /**
+     * Sets whether the rendering operation has been \a stopped and any ongoing
+     * rendering should be canceled immediately.
+     *
+     * \see renderingStopped()
+     */
     void setRenderingStopped( bool stopped ) {mRenderingStopped = stopped;}
 
     /**
@@ -375,18 +474,50 @@ class CORE_EXPORT QgsRenderContext
      */
     void setPainter( QPainter *p ) {mPainter = p;}
 
+    /**
+     * Sets a mask QPainter for the render operation. Ownership of the painter
+     * is not transferred and the QPainter must stay alive for the duration
+     * of any rendering operations.
+     * Multiple mask painters can be defined and the second parameter gives a unique identifier to each one.
+     * \see maskPainter()
+     */
+    void setMaskPainter( QPainter *p, int id = 0 ) { mMaskPainter[id] = p; }
+
+    /**
+     * When rendering a map layer in a second pass (for selective masking),
+     * some symbol layers may be disabled.
+     *
+     * Sets the list of disabled symbol layers.
+     * \see disabledSymbolLayers()
+     * \see isSymbolLayerEnabled()
+     * \since QGIS 3.12
+     */
+    void setDisabledSymbolLayers( const QSet<const QgsSymbolLayer *> &symbolLayers ) { mDisabledSymbolLayers = symbolLayers; }
+
+    /**
+     * Sets whether rendering operations should use vector operations instead
+     * of any faster raster shortcuts.
+     *
+     * \see forceVectorOutput()
+     */
     void setForceVectorOutput( bool force );
 
     /**
      * Assign new labeling engine
      * \note not available in Python bindings
      */
-    void setLabelingEngine( QgsLabelingEngine *engine2 ) { mLabelingEngine = engine2; } SIP_SKIP
+    void setLabelingEngine( QgsLabelingEngine *engine ) { mLabelingEngine = engine; } SIP_SKIP
+
+    /**
+     * Sets the \a color to use when rendering selected features.
+     *
+     * \see selectionColor()
+     */
     void setSelectionColor( const QColor &color ) { mSelectionColor = color; }
 
     /**
      * Sets whether vector selections should be shown in the rendered map
-     * \param showSelection set to true if selections should be shown
+     * \param showSelection set to TRUE if selections should be shown
      * \see showSelection
      * \see setSelectionColor
      * \since QGIS v2.4
@@ -394,14 +525,42 @@ class CORE_EXPORT QgsRenderContext
     void setShowSelection( bool showSelection );
 
     /**
-     * Returns true if the rendering optimization (geometry simplification) can be executed
+     * Returns TRUE if the rendering optimization (geometry simplification) can be executed
+     *
+     * \see setUseRenderingOptimization()
      */
     bool useRenderingOptimization() const;
 
+    /**
+     * Sets whether the rendering optimization (geometry simplification) should be executed
+     *
+     * \see useRenderingOptimization()
+     */
     void setUseRenderingOptimization( bool enabled );
 
-    //! Added in QGIS v2.4
+    /**
+     * Returns the simplification settings to use when rendering vector layers.
+     *
+     * The default is to use no simplification.
+     *
+     * \see setVectorSimplifyMethod()
+     * \since QGIS 2.4
+     */
     const QgsVectorSimplifyMethod &vectorSimplifyMethod() const { return mVectorSimplifyMethod; }
+
+    /**
+     * Sets the simplification setting to use when rendering vector layers.
+     *
+     * This can be used to specify simplification methods to apply during map exports and renders,
+     * e.g. to allow vector layers to be simplified to an appropriate maximum level of detail
+     * during PDF exports or to speed up layer rendering
+     *
+     * The default is to use no simplification.
+     *
+     * \see vectorSimplifyMethod()
+     *
+     * \since QGIS 2.4
+     */
     void setVectorSimplifyMethod( const QgsVectorSimplifyMethod &simplifyMethod ) { mVectorSimplifyMethod = simplifyMethod; }
 
     /**
@@ -452,16 +611,30 @@ class CORE_EXPORT QgsRenderContext
 
     /**
      * Sets the segmentation tolerance applied when rendering curved geometries
-    \param tolerance the segmentation tolerance*/
+     * \param tolerance the segmentation tolerance
+     * \see segmentationTolerance()
+     * \see segmentationToleranceType()
+     */
     void setSegmentationTolerance( double tolerance ) { mSegmentationTolerance = tolerance; }
-    //! Gets the segmentation tolerance applied when rendering curved geometries
+
+    /**
+     * Gets the segmentation tolerance applied when rendering curved geometries
+     * \see setSegmentationTolerance()
+     */
     double segmentationTolerance() const { return mSegmentationTolerance; }
 
     /**
      * Sets segmentation tolerance type (maximum angle or maximum difference between curve and approximation)
-    \param type the segmentation tolerance typename*/
+     * \param type the segmentation tolerance typename
+     * \see segmentationToleranceType()
+     * \see segmentationTolerance()
+     */
     void setSegmentationToleranceType( QgsAbstractGeometry::SegmentationToleranceType type ) { mSegmentationToleranceType = type; }
-    //! Gets segmentation tolerance type (maximum angle or maximum difference between curve and approximation)
+
+    /**
+     * Gets segmentation tolerance type (maximum angle or maximum difference between curve and approximation)
+     * \see setSegmentationToleranceType()
+     */
     QgsAbstractGeometry::SegmentationToleranceType segmentationToleranceType() const { return mSegmentationToleranceType; }
 
     // Conversions
@@ -519,6 +692,100 @@ class CORE_EXPORT QgsRenderContext
       mTextRenderFormat = format;
     }
 
+    /**
+     * Returns the list of rendered feature handlers to use while rendering map layers.
+     * \see hasRenderedFeatureHandlers()
+     * \since QGIS 3.10
+     */
+    QList<QgsRenderedFeatureHandlerInterface *> renderedFeatureHandlers() const;
+
+    /**
+     * Returns TRUE if the context has any rendered feature handlers.
+     * \see renderedFeatureHandlers()
+     * \since QGIS 3.10
+     */
+    bool hasRenderedFeatureHandlers() const { return mHasRenderedFeatureHandlers; }
+
+    /**
+     * Attaches a mask id provider to the context. It will allow some rendering operations to set the current mask id
+     * based on the context (label layer names and label rules for instance).
+     * \see QgsMaskIdProvider
+     * \see setCurrentMaskId()
+     * \see maskIdProvider()
+     * \since QGIS 3.12
+     */
+    void setMaskIdProvider( QgsMaskIdProvider *provider ) { mMaskIdProvider = provider; }
+
+    /**
+     * Returns the mask id provider attached to the context.
+     * \see setMaskIdProvider()
+     * \since QGIS 3.12
+     */
+    const QgsMaskIdProvider *maskIdProvider() const { return mMaskIdProvider; }
+
+    /**
+     * Stores a mask id as the "current" one.
+     * \see currentMaskId()
+     * \since QGIS 3.12
+     */
+    void setCurrentMaskId( int id ) { mCurrentMaskId = id; }
+
+    /**
+     * Returns the current mask id, which can be used with maskPainter()
+     * \see setCurrentMaskId()
+     * \see maskPainter()
+     * \since QGIS 3.12
+     */
+    int currentMaskId() const { return mCurrentMaskId; }
+
+    /**
+     * Sets GUI preview mode.
+     * GUI preview mode is used to change the behavior of some renderings when
+     * they are done to preview of symbology in the GUI.
+     * This is especially used to display mask symbol layers rather than painting them
+     * in a mask painter, which is not meant to be visible, by definition.
+     * \see isGuiPreview
+     * \since QGIS 3.12
+     */
+    void setIsGuiPreview( bool preview ) { mIsGuiPreview = preview; }
+
+    /**
+     * Returns the Gui preview mode.
+     * GUI preview mode is used to change the behavior of some renderings when
+     * they are done to preview of symbology in the GUI.
+     * This is especially used to display mask symbol layers rather than painting them
+     * in a mask painter, which is not meant to be visible, by definition.
+     * \see isGuiPreview
+     * \see setIsGuiPreview
+     * \since QGIS 3.12
+     */
+    bool isGuiPreview() const { return mIsGuiPreview; }
+
+    /**
+     * Gets custom rendering flags. Layers might honour these to alter their rendering.
+     * \returns a map of custom flags
+     * \see setCustomRenderingFlag()
+     * \since QGIS 3.12
+     */
+    QVariantMap customRenderingFlags() const { return mCustomRenderingFlags; }
+
+    /**
+     * Sets a custom rendering flag. Layers might honour these to alter their rendering.
+     * \param flag the flag name
+     * \param value the flag value
+     * \see customRenderingFlags()
+     * \since QGIS 3.12
+     */
+    void setCustomRenderingFlag( const QString &flag, const QVariant &value ) { mCustomRenderingFlags[flag] = value; }
+
+    /**
+     * Clears the specified custom rendering flag.
+     * \param flag the flag name
+     * \see setCustomRenderingFlag()
+     * \since QGIS 3.12
+     */
+    void clearCustomRenderingFlag( const QString &flag ) { mCustomRenderingFlags.remove( flag ); }
+
   private:
 
     Flags mFlags;
@@ -526,6 +793,33 @@ class CORE_EXPORT QgsRenderContext
     //! Painter for rendering operations
     QPainter *mPainter = nullptr;
 
+    /**
+     * Mask painters for selective masking.
+     * Multiple mask painters can be defined for a rendering. The map key is a unique identifier for each mask painter.
+     * \see mMaskIdProvider
+     * \since QGIS 3.12
+     */
+    QMap<int, QPainter *> mMaskPainter;
+
+    /**
+     * Pointer to a mask id provider
+     * \see QgsMaskIdProvider
+     * \since QGIS 3.12
+     */
+    QgsMaskIdProvider *mMaskIdProvider = nullptr;
+
+    /**
+     * Current mask identifier
+     * \since QGIS 3.12
+     */
+    int mCurrentMaskId = -1;
+
+    /**
+     * Whether we are rendering a preview of a symbol / label
+     * \since QGIS 3.12
+     */
+    bool mIsGuiPreview = false;
+
     //! For transformation between coordinate systems. Can be invalid if on-the-fly reprojection is not used
     QgsCoordinateTransform mCoordTransform;
 
@@ -550,7 +844,7 @@ class CORE_EXPORT QgsRenderContext
     //! Map scale
     double mRendererScale = 1.0;
 
-    //! Newer labeling engine implementation (can be nullptr)
+    //! Newer labeling engine implementation (can be NULLPTR)
     QgsLabelingEngine *mLabelingEngine = nullptr;
 
     //! Color used for features that are marked as selected
@@ -577,6 +871,11 @@ class CORE_EXPORT QgsRenderContext
     QgsPathResolver mPathResolver;
 
     TextRenderFormat mTextRenderFormat = TextFormatAlwaysOutlines;
+    QList< QgsRenderedFeatureHandlerInterface * > mRenderedFeatureHandlers;
+    bool mHasRenderedFeatureHandlers = false;
+    QVariantMap mCustomRenderingFlags;
+
+    QSet<const QgsSymbolLayer *> mDisabledSymbolLayers;
 
 #ifdef QGISDEBUG
     bool mHasTransformContext = false;
@@ -585,4 +884,150 @@ class CORE_EXPORT QgsRenderContext
 
 Q_DECLARE_OPERATORS_FOR_FLAGS( QgsRenderContext::Flags )
 
+#ifndef SIP_RUN
+
+/**
+ * \ingroup core
+ *
+ * Scoped object for temporary replacement of a QgsRenderContext destination painter.
+ *
+ * Temporarily swaps out the destination QPainter object for a QgsRenderContext for the lifetime of the object,
+ * before replacing it to the original painter on destruction.
+ *
+ * \note Not available in Python bindings
+ * \since QGIS 3.14
+ */
+class QgsScopedRenderContextPainterSwap
+{
+  public:
+
+    /**
+     * Constructor for QgsScopedRenderContextPainterSwap.
+     *
+     * Swaps the destination painter for \a context (set QgsRenderContext::setPainter() ) to
+     * \a temporaryPainter.
+     */
+    QgsScopedRenderContextPainterSwap( QgsRenderContext &context, QPainter *temporaryPainter )
+      : mContext( context )
+      , mPreviousPainter( context.painter() )
+    {
+      mContext.setPainter( temporaryPainter );
+    }
+
+    /**
+     * Resets the destination painter for the context back to the original QPainter object.
+     */
+    void reset()
+    {
+      if ( !mReleased )
+      {
+        mContext.setPainter( mPreviousPainter );
+        mReleased = true;
+      }
+    }
+
+    /**
+     * Returns the destination painter for the context back to the original QPainter object.
+     */
+    ~QgsScopedRenderContextPainterSwap()
+    {
+      reset();
+    }
+
+  private:
+
+    QgsRenderContext &mContext;
+    QPainter *mPreviousPainter = nullptr;
+    bool mReleased = false;
+};
+
+
+/**
+ * \ingroup core
+ *
+ * Scoped object for temporary scaling of a QgsRenderContext for millimeter based rendering.
+ *
+ * Temporarily scales the destination QPainter for a QgsRenderContext to use millimeter based units for the lifetime of the object,
+ * before returning it to pixel based units on destruction.
+ *
+ * \note Not available in Python bindings
+ * \since QGIS 3.14
+ */
+class QgsScopedRenderContextScaleToMm
+{
+  public:
+
+    /**
+     * Constructor for QgsScopedRenderContextScaleToMm.
+     *
+     * Rescales the destination painter (see QgsRenderContext::painter() ) to use millimeter based units.
+     *
+     * \warning It is the caller's responsibility to ensure that \a context is initially scaled to use pixel based units!
+     */
+    QgsScopedRenderContextScaleToMm( QgsRenderContext &context )
+      : mContext( context )
+    {
+      if ( mContext.painter() )
+        mContext.painter()->scale( mContext.scaleFactor(), mContext.scaleFactor() );
+    }
+
+    /**
+     * Returns the destination painter back to pixel based units.
+     */
+    ~QgsScopedRenderContextScaleToMm()
+    {
+      if ( mContext.painter() )
+        mContext.painter()->scale( 1.0 / mContext.scaleFactor(), 1.0 / mContext.scaleFactor() );
+    }
+
+  private:
+
+    QgsRenderContext &mContext;
+};
+
+
+/**
+ * \ingroup core
+ *
+ * Scoped object for temporary scaling of a QgsRenderContext for pixel based rendering.
+ *
+ * Temporarily scales the destination QPainter for a QgsRenderContext to use pixel based units for the lifetime of the object,
+ * before returning it to millimeter based units on destruction.
+ *
+ * \note Not available in Python bindings
+ * \since QGIS 3.14
+ */
+class QgsScopedRenderContextScaleToPixels
+{
+  public:
+
+    /**
+     * Constructor for QgsScopedRenderContextScaleToPixels.
+     *
+     * Rescales the destination painter (see QgsRenderContext::painter() ) to use pixel based units.
+     *
+     * \warning It is the caller's responsibility to ensure that \a context is initially scaled to use millimeter based units!
+     */
+    QgsScopedRenderContextScaleToPixels( QgsRenderContext &context )
+      : mContext( context )
+    {
+      if ( mContext.painter() )
+        mContext.painter()->scale( 1.0 / mContext.scaleFactor(), 1.0 / mContext.scaleFactor() );
+    }
+
+    /**
+     * Returns the destination painter back to millimeter based units.
+     */
+    ~QgsScopedRenderContextScaleToPixels()
+    {
+      if ( mContext.painter() )
+        mContext.painter()->scale( mContext.scaleFactor(), mContext.scaleFactor() );
+    }
+
+  private:
+
+    QgsRenderContext &mContext;
+};
+#endif
+
 #endif
